package com.vk;

import com.vk.Complex;

public class CMath{
 public static Complex neg(Complex x){
  return new Complex(-x.real(), -x.imag());
 }
 public static Complex conjg(Complex x){
  return new Complex(x.real(), -x.imag());
 }
 public static Complex add(Complex x, double y){
  return new Complex(x.real()+y, x.imag());
 }
 public static Complex mul(Complex x, double y){
  return new Complex(x.real()*y, x.imag()*y);
 }
 public static Complex add(Complex x, Complex y){
  double z1,z2;
  z1=x.real()+y.real(); z2=x.imag()+y.imag();
  return new Complex(z1,z2); 
 } 
 public static Complex sub(Complex x, Complex y){
  double z1,z2;
  z1=x.real()-y.real(); z2=x.imag()-y.imag();
  return new Complex(z1,z2); 
 } 
 public static Complex mul(Complex x, Complex y){
  double x1,x2,y1,y2,z1,z2;
  x1=x.real(); x2=x.imag(); y1=y.real(); y2=y.imag();
  z1=x1*y1-x2*y2; z2=x1*y2+x2*y1;
  return new Complex(z1,z2); 
 } 
 public static Complex div(Complex x, Complex y){
  double x1,x2,y1,y2,z1,z2;
  x1=x.real(); x2=x.imag(); y1=y.real(); y2=y.imag();
  z1=x1*y1+x2*y2; z2=x2*y1-x1*y2; x1=y1*y1+y2*y2; z1 /= x1; z2 /= x1; 
  return new Complex(z1,z2); 
 } 
 public static Complex exp(Complex x){
  double x1,x2,z1,z2; x1=x.real(); x2=x.imag();
  z1=Math.cos(x2); z2=Math.sin(x2); x2=Math.exp(x1); z1 *= x2; z2 *= x2;
  return new Complex(z1,z2); 
 }
 public static Complex cos(Complex x){
  double x1,x2,z1,z2; x1=x.real(); x2=x.imag();
  z1=Math.cos(x1); z2=-Math.sin(x1); x1=Math.exp(x2); z1 *= (x1+1./x1)/2.; z2 *= (x1-1./x1)/2.;
  return new Complex(z1,z2); 
 }
 public static Complex sin(Complex x){
  double x1,x2,z1,z2; x1=x.real(); x2=x.imag();
  z1=Math.sin(x1); z2=Math.cos(x1); x1=Math.exp(x2); z1 *= (x1+1./x1)/2.; z2 *= (x1-1./x1)/2.;
  return new Complex(z1,z2); 
 }
 public static Complex tan(Complex x){
  return div(sin(x),cos(x)); 
 }
 public static Complex pow(Complex x, double y){
  double x1,x2,z1,z2; x1=Math.pow(x.abs(),y); x2=x.phase()*y;
  z1=x1*Math.cos(x2); z2=x1*Math.sin(x2);
  return new Complex(z1,z2); 
 }
 public static Complex log(Complex x){
  double z1,z2; z1=Math.log(x.abs()); z2=x.phase();
  return new Complex(z1,z2); 
 }
 // procedure fq = fft(fx,k) was translated from a similar fortran procedure
 // obtained from Pavel Petrashen. It calculates
 //     fq[n] = sum{m = 0,...,N-1} exp( sign(k)*2*pi*i*n*m/N )*fx[m] 
 // where N = 2^abs(k) with k to be integer, n = 0,...,N-1
 //
 public static Complex[] fft(Complex[] car, int k){
  int i,l,n,n1,n2,m,m1,ks,ns;
  ks = Math.abs(k); ns =  1; for( i=0; i < ks; i++ ){ ns *= 2; }
  Complex[] ca = new Complex[ns]; for( i=0; i < ns; i++ ){ ca[i]=car[i]; }
  Complex[] cb = new Complex[ns]; Complex[] cc = new Complex[ns/2];
  Complex c; int[] inv = new int[ns]; inv[0]=1; n2=1;
  for( l=0; l < ks; l++ ){
   for( i=0; i < n2; i++ ){
    n=i+n2; inv[n]=inv[i]*2; inv[i]=inv[n]-1;
   } n2 *= 2;
  }
  c = exp(new Complex(0.,2.*Math.PI/ns)); n2 /= 2;
  cc[0]=c; for( i=1; i < n2; i++ ){ cc[i]=mul(cc[i-1],c); } 
  for( i=0; i < ns; i++ ){ m=inv[i]-1; cb[m]=ca[i]; }
  for( i=0; i < n2; i++ ){
   m=i*2+1; l=m-1; ca[i]=add(cb[l],cb[m]); ca[i+n2]=sub(cb[l],cb[m]);
  }
  m1=n2;
  for( i=1; i < ks; i++ ){
   m=m1+2; m1 /= 2; 
   for( n=m; n <= ns; n += 2){
    l=(n-1)/(m-2); l *= m1; c=cc[l-1]; if( k < 0 ){ c=conjg(c); }
    n1=n-1; ca[n1] = mul(ca[n1],c);
   } 
   for( l=0; l < ns; l++ ){ cb[l]=ca[l]; }
   for( n=0; n < n2; n++ ){
    m=n*2+1; l=m-1; ca[n]=add(cb[l],cb[m]); ca[n+n2]=sub(cb[l],cb[m]);
   }
  } return ca;
 } 
 public static Complex bj0(Complex arg){
  Complex zsq, argsq, n, d, pzero, qzero, c0, c1, o;   int i;
  double tpi = 0.6366197723675813430755350535;
  double pio4 = Math.PI/4; c0 = new Complex(0.,0.); c1 = new Complex(1.,0.); 
  double[]  p1= {
   0.4933787251794133561816813446e21, -0.1179157629107610536038440800e21,
   0.6382059341072356562289432465e19, -0.1367620353088171386865416609e18,
   0.1434354939140344111664316553e16, -0.8085222034853793871199468171e13,
   0.2507158285536881945555156435e11, -0.4050412371833132706360663322e8,
   0.2685786856980014981415848441e5 };
  double[]  q1= {
   0.4933787251794133562113278438e21, 0.5428918384092285160200195092e19,
   0.3024635616709462698627330784e17, 0.1127756739679798507056031594e15,
   0.3123043114941213172572469442e12, 0.6699987672982239671814028660e9,
   0.1114636098462985378182402543e7, 0.1363063652328970604442810507e4, 1.0 };
  double[] p2 = {
   0.5393485083869438325262122897e7, 0.1233238476817638145232406055e8,
   0.8413041456550439208464315611e7, 0.2016135283049983642487182349e7,
   0.1539826532623911470917825993e6, 0.2485271928957404011288128951e4, 0.0 };
  double[] q2 = {
   0.5393485083869438325560444960e7, 0.1233831022786324960844856182e8,
   0.8426449050629797331554404810e7, 0.2025066801570134013891035236e7,
   0.1560017276940030940592769933e6, 0.2615700736920839685159081813e4, 1.0};
  double[] p3 = {
   -0.3984617357595222463506790588e4, -0.1038141698748464093880530341e5,
   -0.8239066313485606568803548860e4, -0.2365956170779108192723612816e4,
   -0.2262630641933704113967255053e3, -0.4887199395841261531199129300e1, 0.0 };
  double[] q3 = {
   0.2550155108860942382983170882e6, 0.6667454239319826986004038103e6,
   0.5332913634216897168722255057e6, 0.1560213206679291652539287109e6,
   0.1570489191515395519392882766e5, 0.4087714673983499223402830260e3, 1.0 };
  if(arg.abs() > 8.){ zsq=mul(arg,arg); o=mul(c1,64.); zsq = div(o,zsq); n=c0; d=c0;
   for(i=6; i >= 0; i--){ n=mul(n,zsq); n=add(n,p2[i]); d=mul(d,zsq); d=add(d,q2[i]); }
   pzero=div(n,d); n=c0; d=c0; 
   for(i=6; i >= 0; i--){ n=mul(n,zsq); n=add(n,p3[i]); d=mul(d,zsq); d=add(d,q3[i]); } 
   o=div(n,d); o=mul(o,8.); qzero=div(o,arg); o=mul(c1,pio4); n=sub(arg,o);
   o=mul(c1,tpi); o=div(o,arg); o=pow(o,0.5); d=sin(n); d=mul(qzero,d); n=cos(n); n=mul(pzero,n);
   d=sub(n,d); o=mul(o,d); return(o);
  }
  argsq=mul(arg,arg); n=c0; d=c0;
  for(i=8; i >= 0; i--){ n=mul(argsq,n); n=add(n,p1[i]); d=mul(argsq,d); d=add(d,q1[i]);}
  o=div(n,d); return(o);
}
 public static Complex bj1(Complex arg){
  Complex zsq, xsq, n, d, x, pzero, qzero, c0, c1, o;  int i;
  double tpi = 0.6366197723675813430755350535;
  double pio4 = Math.PI/4; c0 = new Complex(0.,0.); c1 = new Complex(1.,0.); 
  double[] p1 = {
   0.581199354001606143928050809e21, -0.6672106568924916298020941484e20,
   0.2316433580634002297931815435e19, -0.3588817569910106050743641413e17,
   0.2908795263834775409737601689e15, -0.1322983480332126453125473247e13,
   0.3413234182301700539091292655e10, -0.4695753530642995859767162166e7,
   0.2701122710892323414856790990e4 };
  double[] q1 = {
   0.1162398708003212287858529400e22, 0.1185770712190320999837113348e20,
   0.6092061398917521746105196863e17, 0.2081661221307607351240184229e15,
   0.5243710262167649715406728642e12, 0.1013863514358673989967045588e10,
   0.1501793594998585505921097578e7, 0.1606931573481487801970916749e4, 1.0 };
  double[] p2 = {
   -0.4435757816794127857114720794e7, -0.9942246505077641195658377899e7,
   -0.6603373248364939109255245434e7, -0.1523529351181137383255105722e7,
   -0.1098240554345934672737413139e6, -0.1611616644324610116477412898e4, 0.0 };
  double[] q2 = {
   -0.4435757816794127856828016962e7, -0.9934124389934585658967556309e7,
   -0.6585339479723087072826915069e7, -0.1511809506634160881644546358e7,
   -0.1072638599110382011903063867e6, -0.1455009440190496182453565068e4, 1.0 };
  double[] p3 = {
   0.3322091340985722351859704442e5, 0.8514516067533570196555001171e5,
   0.6617883658127083517939992166e5, 0.1849426287322386679652009819e5,
   0.1706375429020768002061283546e4, 0.3526513384663603218592175580e2, 0.0 };
  double[] q3 = {
   0.7087128194102874357377502472e6, 0.1819458042243997298924553839e7,
   0.1419460669603720892855755253e7, 0.4002944358226697511708610813e6,
   0.3789022974577220264142952256e5, 0.8638367769604990967475517183e3, 1.0 };
  x = arg;
  if( x.abs() > 8. ) {  zsq=mul(arg,arg); o=mul(c1,64.); zsq = div(o,zsq); n=c0; d=c0;
   for(i=6; i >=0 ; i--){n=mul(n,zsq); n=add(n,p2[i]); d=mul(d,zsq); d=add(d,q2[i]); }
   pzero=div(n,d); n=c0; d=c0; 
   for(i=6; i >= 0; i--){ n=mul(n,zsq); n=add(n,p3[i]); d=mul(d,zsq); d=add(d,q3[i]); } 
   o=div(n,d); o=mul(o,8.); qzero=div(o,arg); o=mul(c1,3.*pio4); n=sub(x,o);
   o=mul(c1,tpi); o=div(o,x); o=pow(o,0.5); d=sin(n); d=mul(qzero,d); n=cos(n); n=mul(pzero,n);
   d=sub(n,d); o=mul(o,d); return(o);
  }
  xsq=mul(x,x); n=c0; d=c0;
  for(i=8; i >= 0; i--){ n=mul(xsq,n); n=add(n,p1[i]); d=mul(xsq,d); d=add(d,q1[i]);}
  o=div(n,d); o=mul(arg,o); return(o);
 } 
}
